import 'dart:convert';

import 'package:dartrofit/dartrofit.dart';
import 'package:dartrofit/src/annotations/header.dart';
import 'package:dartrofit/src/annotations/multipart.dart';
import 'package:dartrofit/src/annotations/querys.dart';
import 'package:dartrofit/src/request_builder.dart';
import 'package:http/http.dart';

abstract class ParameterHandler<T> {
  void apply(RequestBuilder builder, T value);
}

/// [Query]
class QueryHandler extends ParameterHandler<dynamic> {
  final String name;

  QueryHandler(this.name) : assert(name != null);

  @override
  void apply(RequestBuilder builder, value) {
    if (value == null) return; // skip null value
    builder.addQueryParam(name, value.toString());
  }
}

/// [QueryMap]
class QueryMapHandler extends ParameterHandler<Map<String, dynamic>> {
  @override
  void apply(RequestBuilder builder, Map<String, dynamic> value) {
    if (value == null) {
      throw ArgumentError('Query map was null');
    }

    value.forEach((entryKey, entryValue) {
      if (entryKey == null) {
        throw ArgumentError('Query map contained null key');
      }

      if (entryValue == null) {
        throw ArgumentError(
            "Query map contained null value for key '$entryKey'");
      }

      builder.addQueryParam(entryKey, entryValue.toString());
    });
  }
}

/// [Body]
class BodyHandler extends ParameterHandler<dynamic> {
  final String encodeName;

  BodyHandler(this.encodeName) : assert(encodeName != null);

  @override
  void apply(RequestBuilder builder, value) {
    if (value == null) return;
    builder.body = value;
    builder.encoding = Encoding.getByName(encodeName);
  }
}

/// [Headers]
class HeadersHandler extends ParameterHandler<List<String>> {
  @override
  void apply(RequestBuilder builder, List<String> value) {
    if (value == null || value.isEmpty) return;
    for (var str in value) {
      // expect 'name: value'
      if (str.contains(':')) {
        final split = str.split(':');
        final name = split[0].trim();
        final value = split[1].trim();
        if (!Strings.isNullOrEmpty(name) && !Strings.isNullOrEmpty(value)) {
          builder.headers[name] = value;
        }
      }
    }
  }
}

/// [Header]
class HeaderHandler extends ParameterHandler<dynamic> {
  final String name;

  HeaderHandler(this.name) : assert(name != null);

  @override
  void apply(RequestBuilder builder, value) {
    if (value == null) return;
    builder.headers[name] = value.toString();
  }
}

/// [HeaderMap]
class HeaderMapHandler extends ParameterHandler<dynamic> {
  @override
  void apply(RequestBuilder builder, value) {
    if (value == null) return;
    if (value is Map) {
      final casted = value.cast<String, String>();
      builder.headers.addAll(casted);
    }
  }
}

/// [PartField]
class PartFieldHandler extends ParameterHandler<dynamic> {
  final String name;

  PartFieldHandler(this.name) : assert(name != null);

  @override
  void apply(RequestBuilder builder, value) {
    if (value == null) return;
    builder.multipartFields[name] = value.toString();
  }
}

/// [PartFieldMap]
class PartFieldMapHandler extends ParameterHandler<dynamic> {
  @override
  void apply(RequestBuilder builder, value) {
    if (value == null) return;
    if (value is Map) {
      final casted = value.cast<String, String>();
      if (casted.isNotEmpty) {
        builder.multipartFields.addAll(casted);
      }
    }
  }
}

/// [PartFile]
class PartFileHandler extends ParameterHandler<dynamic> {
  @override
  void apply(RequestBuilder builder, value) {
    if (value == null) return;
    if (value is MultipartFile) {
      builder.multipartFiles.add(value);
    }
  }
}

/// [PartFileList]
class PartFileListHandler extends ParameterHandler<dynamic> {
  @override
  void apply(RequestBuilder builder, value) {
    if (value == null) return;
    if (value is Iterable) {
      final casted = value.cast<MultipartFile>();
      builder.multipartFiles.addAll(casted);
    }
  }
}
